﻿using System;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Linq;

  using Common;
  using Configuration;
  using Core;
  using Data.Pool;
  using HeuristicLab.Data;
  using HeuristicLab.Encodings.IntegerVectorEncoding;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Analyzer;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Individual;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Problem.BenchmarkSuite;

  using Interpreter;
  using Optimization;
  using Parameters;
  using Persistence.Default.CompositeSerializers.Storable;
  using Random;

  [StorableClass]
  public abstract class PushProblem : SingleObjectiveBasicProblem<IntegerVectorEncoding> {
    [Storable]
    protected readonly PushConfigurationParameterCollection config;
    protected PushInterpreterPool pool;
    protected readonly ObjectPool<IRandom> randomPool = new ObjectPool<IRandom>(() => new MersenneTwister(), Environment.ProcessorCount * 4);

    [Storable]
    protected readonly IPushEvaluator PushEvaluator;

    private const string BestTrainingSolutionResultName = "Best Solution";
    private const string TestQualityResultName = "Test Quality";

    protected PushProblem(PushBenchmarkSuiteEvaluator evaluator) {
      config = new PushConfigurationParameterCollection();
      PushEvaluator = evaluator;

      InitData();
      InitEvents();
      InitParameters();
      InitEncoding();
      InitOperators();
    }

    [StorableConstructor]
    protected PushProblem(bool deserializing)
      : base(deserializing) {
    }

    protected PushProblem(PushProblem original, Cloner cloner)
      : base(original, cloner) {
      config = cloner.Clone(original.config);
      PushEvaluator = cloner.Clone(original.PushEvaluator);

      InitData();
      InitEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitData();
      InitEvents();
      InitParameters();
    }

    private void InitData() {
      pool = new PushInterpreterPool(Environment.ProcessorCount * 2, 4096, 1024, config);
    }

    private void InitEvents() {
      config.EnabledExpressionsChanged += EnabledExpressionsChanged;
    }

    private void EnabledExpressionsChanged(object sender, EnabledExpressionsChangedEventArgs e) {
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.BoundsParameter.Value[0, 1] = config.EnabledExpressions.Count;
    }

    public ILookupParameter<IntValue> SeedParamater
    {
      get { return (ILookupParameter<IntValue>)Parameters["Seed"]; }
    }

    #region Parameters
    private const string InitProgramLengthParameterName = "InitProgramLength";
    private const string InitProgramLengthParameterDescription = "This is the initial size of a push program.";
    private const string PushConfigurationParameterName = "PushConfiguration";

    public const string CasesScopeParameterName = "CaseQualities";
    public const string CaseQualitiesScopeParameterName = "CaseQualities";

    private void InitParameters() {
      foreach (var paramater in config.Parameters) {
        if (!Parameters.ContainsKey(paramater.Name)) {
          Parameters.Add(paramater);
        }
      }

      if (!Parameters.ContainsKey(PushConfigurationParameterName))
        Parameters.Add(new ValueParameter<IReadOnlyPushConfiguration>(PushConfigurationParameterName, config) {
          Hidden = true
        });

      if (!Parameters.ContainsKey(InitProgramLengthParameterName)) {
        Parameters.Add(new FixedValueParameter<IntValue>(
          InitProgramLengthParameterName,
          InitProgramLengthParameterDescription,
          new IntValue(50)));
      }

      Encoding.LengthParameter = InitProgramLengthParameter as IFixedValueParameter<IntValue>;

      if (!Parameters.ContainsKey(CasesScopeParameterName))
        Parameters.Add(new LookupParameter<BoolArray>(CasesScopeParameterName, "The training cases that have been successfully executed."));

      if (!Parameters.ContainsKey(CaseQualitiesScopeParameterName))
        Parameters.Add(new LookupParameter<DoubleArray>(CaseQualitiesScopeParameterName, "The quality of every single training case for each individual"));
    }

    protected override void OnReset() {
      base.OnReset();

      // clear pools and free reserved memory
      pool.Clear();
      IndividualMapper.Clear();
    }

    private void InitEncoding() {
      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;
      Encoding.Length = config.MaxPointsInProgram;
    }

    private void InitOperators() {
      Operators.Add(new PushExpressionFrequencyAnalyzer());
    }

    /// <summary>
    ///     This is the inital size of an push program generated by a solution creator
    /// </summary>
    public IValueParameter<IntValue> InitProgramLengthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[InitProgramLengthParameterName]; }
    }

    public int InitProgramLength
    {
      get { return InitProgramLengthParameter.Value.Value; }
      set
      {
        InitProgramLengthParameter.Value.Value = value;
      }
    }
    #endregion

    public override bool Maximization { get { return false; } }

    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      IndividualMapper.Reset();

      var bestQuality = Maximization ? qualities.Max() : qualities.Min();
      var bestIdx = Array.IndexOf(qualities, bestQuality);
      var vector = individuals[bestIdx].IntegerVector();
      var seed = vector.GetSeed();
      var rand = randomPool.Allocate();
      rand.Reset(seed);

      var program = (PushProgram)vector.ToPushProgram(config, rand).Clone();
      var isIndividualBetter = AnalyzeBestTrainingSolution(program, bestQuality, results, rand);

      if (isIndividualBetter) {
        rand.Reset(seed);
        AnalyzeBestTestSolution(program, results, rand);
      }

      randomPool.Free(rand);
    }

    private void AnalyzeBestTestSolution(PushProgram program, ResultCollection results, IRandom random) {
      var testResult = PushEvaluator.EvaluateTraining(pool, program, random);

      if (!results.ContainsKey(TestQualityResultName)) {
        results.Add(new Result(TestQualityResultName, new DoubleValue(testResult.AvgQuality)));
      } else {
        ((DoubleValue)results[TestQualityResultName].Value).Value = testResult.AvgQuality;
      }
    }

    private bool AnalyzeBestTrainingSolution(PushProgram program, double bestQuality, ResultCollection results, IRandom random) {
      if (!results.ContainsKey(BestTrainingSolutionResultName)) {
        var solution = CreatePushSolution(
            program,
            bestQuality,
            random,
            config,
            PushEvaluator);

        results.Add(new Result(BestTrainingSolutionResultName, solution));
        return true;
      }

      var currentBestQuality = ((PushSolution)results[BestTrainingSolutionResultName].Value).Quality;

      if ((!Maximization && currentBestQuality > bestQuality) ||
           (Maximization && currentBestQuality < bestQuality)) {
        results[BestTrainingSolutionResultName].Value = CreatePushSolution(
            program,
            bestQuality,
            random,
            config,
            PushEvaluator);
        return true;
      }

      return false;
    }

    private EvaluationResult EvaluateIntegerVector(IntegerVector vector) {
      var rand = randomPool.Allocate();
      var seed = vector.GetSeed();
      rand.Reset(seed);

      var program = vector.ToPushProgram(config, rand);
      var result = PushEvaluator.EvaluateTraining(pool, program, rand);
      randomPool.Free(rand);

      return result;
    }

    protected abstract PushSolution CreatePushSolution(
      PushProgram program,
      double bestQuality,
      IRandom random,
      IReadOnlyPushConfiguration config,
      IPushEvaluator evaluator);

    public override double Evaluate(Individual individual, IRandom random) {
      var vector = individual.IntegerVector();
      var result = EvaluateIntegerVector(vector);

      individual[CaseQualitiesScopeParameterName] = new DoubleArray(result.ExampleQualities);

      return result.AvgQuality;
    }
  }
}